import typing as T
from collections import defaultdict
from collections.abc import Mapping

from conllu.exceptions import ParseException
from conllu.serializer import serialize

DEFAULT_EXCLUDE_FIELDS = ('id', 'deprel', 'xpos', 'feats', 'head', 'deps', 'misc')

if T.TYPE_CHECKING:
    class SupportsIndex(T.Protocol):
        def __index__(self) -> int:
            pass

    class SupportsNext(T.Protocol):
        def __next__(self) -> int:
            pass

class Metadata(dict):
    pass

class Token(dict):
    MAPPING = {
        "upostag": "upos",
        "xpostag": "xpos",
        "upos": "upostag",
        "xpos": "xpostag",
    }

    def get(self, key: str, default: T.Optional[T.Any] = None) -> T.Any:
        if key not in self and key in self.MAPPING:
            key = self.MAPPING[key]

        return super(Token, self).get(key, default)

    def __missing__(self, key: str) -> T.Any:
        if key in self.MAPPING:
            return self.get(self.MAPPING[key])

        raise KeyError("'" + key + "'")

    def __str__(self) -> str:
        if 'form' in self:
            return self['form']

        if 'id' in self:
            return f"id={self['id']}"

        return ''

class TokenList(T.List[Token]):
    def __init__(
        self,
        tokens: T.Optional[T.Iterable[Token]] = None,
        metadata: T.Optional[Metadata] = None,
        default_fields: T.Optional[T.Iterable[str]] = None,
    ):
        tokens = tokens or []

        if not isinstance(tokens, list):
            raise ParseException("Can't create TokenList, tokens is not a list.")

        if len(tokens) > 0 and not isinstance(tokens[0], Token):
            tokens = [Token(token) for token in tokens]

        super(TokenList, self).__init__(tokens)

        self.metadata = metadata or Metadata()
        self.default_fields = default_fields

    def __repr__(self) -> str:
        tokens = ', '.join(str(token) for token in self)
        if not self.metadata:
            return f'TokenList<{tokens}>'
        else:
            metadata = ', '.join(f"{key}: \"{value}\"" for key, value in self.metadata.items())
            return f'TokenList<{tokens}, metadata={{{metadata}}}>'

    def __eq__(self, other: T.Any) -> bool:
        if not isinstance(other, TokenList):
            other = TokenList(other)

        return super(TokenList, self).__eq__(other) and self.metadata == other.metadata

    def __ne__(self, other: T.Any) -> bool:
        return not self == other

    def clear(self) -> None:
        super(TokenList, self).clear()
        self.metadata = Metadata()
        self.default_fields = None

    def copy(self) -> 'TokenList':
        tokens_copy = super().copy()
        return TokenList(tokens_copy, self.metadata, self.default_fields)

    def extend(self, iterable: T.Union['TokenList', T.Iterable[Token]]) -> None:
        if not isinstance(iterable, TokenList):
            iterable = TokenList(iterable)

        super(TokenList, self).extend(iterable)

        self.metadata.update(iterable.metadata)

    def _dict_to_token_and_set_defaults(self, token: T.Union[dict, Token]) -> Token:
        if not isinstance(token, Token):
            token = Token(token)

        if self.default_fields:
            for field in self.default_fields:
                if field not in token:
                    token[field] = "_"

        return token

    def append(self, token: T.Union[dict, Token]) -> None:
        token = self._dict_to_token_and_set_defaults(token)
        super(TokenList, self).append(token)

    def insert(self, i: 'SupportsIndex', token: T.Union[dict, Token]) -> None:
        token = self._dict_to_token_and_set_defaults(token)
        super(TokenList, self).insert(i, token)

    @T.overload
    def __setitem__(self, key: 'SupportsIndex', tokens: T.Union[dict, Token]) -> None: ...  # noqa, pragma: no cover

    @T.overload
    def __setitem__(self, key: slice, tokens: T.Union[T.Iterable[T.Union[dict, Token]], 'TokenList']) -> None: ...  # noqa, pragma: no cover

    def __setitem__(self, key, tokens):  # noqa: F811
        if isinstance(key, slice):
            tokens = [self._dict_to_token_and_set_defaults(token) for token in tokens]
            super(TokenList, self).__setitem__(key, tokens)
        else:
            token = tokens
            token = self._dict_to_token_and_set_defaults(token)
            super(TokenList, self).__setitem__(key, token)

    def serialize(self) -> str:
        return serialize(self)

    @staticmethod
    def head_to_token(sentence: 'TokenList') -> T.Dict[int, T.List[Token]]:
        if not sentence:
            raise ParseException("Can't parse tree, need a tokenlist as input.")

        if "head" not in sentence[0]:
            raise ParseException("Can't parse tree, missing 'head' field.")

        head_indexed = defaultdict(list)
        for token in sentence:
            # Filter out range and decimal ID:s before building tree
            if "id" in token and not isinstance(token["id"], int):
                continue

            # Filter out tokens with negative head, they are sometimes used to
            # specify tokens which should not be included in tree
            # Also filter out those that have no head, just exclude them from the tree.
            if (token.get("head") is None) or token["head"] < 0:
                continue

            head_indexed[token["head"]].append(token)

        if len(head_indexed[0]) == 0:
            raise ParseException("Found no head node, can't build tree")

        return head_indexed

    def to_tree(self) -> 'TokenTree':
        def _create_tree(head_to_token_mapping: T.Dict[int, T.List[Token]], id_: int = 0) -> T.List['TokenTree']:
            return [
                TokenTree(child, _create_tree(head_to_token_mapping, child["id"]))
                for child in head_to_token_mapping[id_]
            ]

        head_indexed = self.head_to_token(self)
        if len(head_indexed[0]) > 1:
            # Introduce fake root node that multiple root nodes can have a single parent
            head_indexed[-1] = [Token(
                [("id", 0), ("form", "_"), ("deprel", "root")]
            )]
            root = _create_tree(head_indexed, -1)[0]
        else:
            root = _create_tree(head_indexed, 0)[0]

        root.set_metadata(self.metadata)
        return root

    def filter(self, **kwargs: T.Any) -> 'TokenList':
        tokens: T.Iterable[Token] = self.copy()

        for query, value in kwargs.items():
            filtered_tokens = []
            for token in tokens:
                if callable(value) and value(traverse_dict(token, query)) is True:
                    filtered_tokens.append(token)
                else:
                    if traverse_dict(token, query) == value:
                        filtered_tokens.append(token)

            tokens = filtered_tokens

        return TokenList(tokens)


_T = T.TypeVar("_T")
def traverse_dict(obj: T.Mapping[str, _T], query: str) -> T.Optional[_T]:
    """
        Get elements inside a nested dict, based on a dict query. The query is defined by a
        string separated by '__'. traverse_dict(foo, 'a__b__c') is roughly equivalent to foo[a][b][c] but
        will short circuit to return None if something on the query is None.
    """
    query_split = query.split('__')

    cur_obj: T.Optional[T.Union[_T, T.Mapping[str, _T]]] = obj
    for name in query_split:
        assert isinstance(cur_obj, Mapping)  # help mypy
        cur_obj = cur_obj.get(name, None)
        if cur_obj is None:
            return None
    assert not isinstance(cur_obj, Mapping)  # help mypy
    return cur_obj


class TokenTree(object):
    token: Token
    children: T.List['TokenTree']
    metadata: T.Optional[Metadata]

    def __init__(self, token: Token, children: T.List['TokenTree'], metadata: T.Optional[Metadata] = None):
        self.token = token
        self.children = children
        self.metadata = metadata

    def set_metadata(self, metadata: T.Optional[Metadata]) -> None:
        self.metadata = metadata

    def __repr__(self) -> str:
        return 'TokenTree<' + \
            'token={id=' + str(self.token['id']) + ', form=' + str(self.token['form']) + '}, ' + \
            'children=' + ('[...]' if self.children else 'None') + \
            '>'

    def __eq__(self, other: T.Any) -> bool:
        if isinstance(other, TokenTree):
            return self.token == other.token and self.children == other.children \
                and self.metadata == other.metadata
        return False

    def to_list(self):
        def _to_list(root_token: TokenTree, token_list: T.List[Token] = []) -> T.List[Token]:
            token_list.append(root_token.token)

            for child_token in root_token.children:
                _to_list(child_token, token_list)

            return token_list

        if not self.token or "id" not in self.token:
            raise ParseException("Could not flatten tree; missing 'id' field.")

        token_list = _to_list(self)
        token_list = sorted(token_list, key=lambda t: t['id'])
        token_list = TokenList(token_list, self.metadata)

        return token_list

    def serialize(self) -> str:
        return serialize(self.to_list())

    def print_tree(self, depth: int = 0, indent: int = 4,
                   exclude_fields: T.Sequence[str] = DEFAULT_EXCLUDE_FIELDS) -> None:
        if not self.token:
            raise ParseException("Can't print, token is None.")

        if "deprel" not in self.token or "id" not in self.token:
            raise ParseException("Can't print, token is missing either the id or deprel fields.")

        relevant_data = self.token.copy()
        for key in exclude_fields:
            if key in relevant_data:
                del relevant_data[key]

        node_repr = ' '.join([
            '{key}:{value}'.format(key=key, value=value)
            for key, value in relevant_data.items()
        ])

        print(' ' * indent * depth + '(deprel:{deprel}) {node_repr} [{idx}]'.format(
            deprel=self.token['deprel'],
            node_repr=node_repr,
            idx=self.token['id'],
        ))
        for child in self.children:
            child.print_tree(depth=depth + 1, indent=indent, exclude_fields=exclude_fields)

class SentenceList(T.List[TokenList]):
    def __init__(
        self,
        sentences: T.Optional[T.Iterable[TokenList]] = None,
        metadata: T.Optional[Metadata] = None,
    ):
        sentences = sentences or []

        if hasattr(sentences, "__next__"):
            sentences = list(sentences)

        if not isinstance(sentences, list):
            raise ParseException("Can't create SentenceList, sentences is not a list.")

        super(SentenceList, self).__init__(sentences)

        self.metadata = metadata or Metadata()

    def __repr__(self) -> str:
        sentences = ', '.join(str(sentence) for sentence in self)
        if not self.metadata:
            return f'[{sentences}]'
        else:
            metadata = ', '.join(f"{key}: \"{value}\"" for key, value in self.metadata.items())
            return f'SentenceList<{sentences}, metadata={{{metadata}}}>'

    def __eq__(self, other: T.Any) -> bool:
        if not isinstance(other, SentenceList):
            other = SentenceList(other)

        return super(SentenceList, self).__eq__(other) and self.metadata == other.metadata

    def __ne__(self, other: T.Any) -> bool:
        return not self == other

    @T.overload
    def __getitem__(self, key: 'SupportsIndex') -> TokenList: ...  # noqa, pragma: no cover

    @T.overload
    def __getitem__(self, key: slice) -> "SentenceList": ...  # noqa, pragma: no cover

    def __getitem__(self, key):  # noqa: F811
        if isinstance(key, slice):
            return SentenceList(
                sentences=super(SentenceList, self).__getitem__(key),
                metadata=self.metadata
            )
        else:
            return super(SentenceList, self).__getitem__(key)

    def clear(self) -> None:
        super(SentenceList, self).clear()
        self.metadata = Metadata()

    def copy(self) -> 'SentenceList':
        sentences_copy = super().copy()
        return SentenceList(sentences_copy, self.metadata)

    def extend(self, iterable: T.Union['SentenceList', T.Iterable[TokenList]]) -> None:
        if not isinstance(iterable, SentenceList):
            iterable = SentenceList(iterable)

        super(SentenceList, self).extend(iterable)

        self.metadata.update(iterable.metadata)


class SentenceGenerator(T.Iterable[TokenList]):
    def __init__(
        self,
        sentences: T.Iterator[TokenList],
        metadata: T.Optional[Metadata] = None,
    ):
        if isinstance(sentences, T.List):
            sentences = iter(sentences)

        if not isinstance(sentences, T.Iterator):
            raise ParseException("Can't create SentenceGenerator, sentences is not an iterator.")

        self.sentences = sentences
        self.metadata = metadata or Metadata()

    def __iter__(self) -> T.Iterator[TokenList]:
        return self.sentences.__iter__()

    def __next__(self) -> TokenList:
        return self.sentences.__next__()

    def __repr__(self) -> str:
        if not self.metadata:
            return f'SentenceGenerator<{id(self)}>'
        else:
            metadata = ', '.join(f"{key}: \"{value}\"" for key, value in self.metadata.items())
            return f'SentenceGenerator<{id(self)}, metadata={{{metadata}}}>'



import contextlib
import doctest
import os
import re
import tempfile
from doctest import OutputChecker
from pathlib import Path
from unittest import mock


class ReadmeOutputChecker(OutputChecker):
    def check_output(self, want, got, optionflags):
        # Allow dicts and lists to be formatted with whitespace around brackets
        if want.startswith("{") or want.startswith("["):
            want = re.sub(r"\s*([\{\[\]\}])\s*", r"\1", want, flags=re.MULTILINE)
            got = re.sub(r"\s*([\{\[\]\}])\s*", r"\1", got, flags=re.MULTILINE)

        return OutputChecker.check_output(self, want, got, optionflags)

@contextlib.contextmanager
def temporary_chdir():
    try:
        old = os.getcwd()
        with tempfile.TemporaryDirectory() as tempdir:
            os.chdir(tempdir)
            yield
    finally:
        os.chdir(old)

def test_readme():
    readme_file = Path("README.md")

    # Copy contents of README, to remove ``` and run all code in one session
    readme_data = ""
    with open(readme_file, "r") as f:
        for line in f:
            if line == "```\n":
                readme_data += "\n"
            else:
                readme_data += line

    with temporary_chdir():
        with mock.patch('doctest.OutputChecker', ReadmeOutputChecker):
            doctest.run_docstring_examples(
                readme_data,
                globs={},
                name=readme_file.name,
                optionflags=(
                    doctest.ELLIPSIS
                    | doctest.NORMALIZE_WHITESPACE
                    | doctest.FAIL_FAST
                    | doctest.REPORT_NDIFF
                ),
            )



import typing as T
from io import StringIO

from conllu.models import Metadata, SentenceGenerator, SentenceList, Token, TokenList, TokenTree
from conllu.parser import (
    _FieldParserType, _MetadataParserType, parse_conllu_plus_fields, parse_sentences, parse_token_and_metadata,
)

__all__ = [
    "parse", "parse_incr", "parse_tree", "parse_tree_incr",
    "SentenceGenerator", "SentenceList", "TokenList", "TokenTree", "Token", "Metadata",
    "parse_conllu_plus_fields", "parse_sentences", "parse_token_and_metadata",
]

def parse(data: str, fields: T.Optional[T.Sequence[str]] = None,
          field_parsers: T.Optional[T.Dict[str, _FieldParserType]] = None,
          metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
          ) -> SentenceList:
    return SentenceList(parse_incr(
        StringIO(data),
        fields=fields,
        field_parsers=field_parsers,
        metadata_parsers=metadata_parsers
    ))

def parse_incr(in_file: T.TextIO, fields: T.Optional[T.Sequence[str]] = None,
               field_parsers: T.Optional[T.Dict[str, _FieldParserType]] = None,
               metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
               ) -> SentenceGenerator:

    if not hasattr(in_file, 'read'):
        raise FileNotFoundError("Invalid file, 'parse_incr' needs an opened file as input")

    if not fields:
        fields = parse_conllu_plus_fields(in_file, metadata_parsers=metadata_parsers)

    def generator():
        for sentence in parse_sentences(in_file):
            yield parse_token_and_metadata(
                sentence,
                fields=fields,
                field_parsers=field_parsers,
                metadata_parsers=metadata_parsers
            )

    return SentenceGenerator(generator())

def parse_tree(data: str) -> T.List[TokenTree]:
    return list(parse_tree_incr(StringIO(data)))

def parse_tree_incr(in_file: T.TextIO) -> T.Iterator[TokenTree]:
    for tokenlist in parse_incr(in_file):
        yield tokenlist.to_tree()



import typing as T

from conllu.exceptions import ParseException

if T.TYPE_CHECKING:
    from conllu.models import TokenList


SERIALIZABLE_TERMINAL_VALUE_TYPES = (int, float, bool, str)


def serialize_field(field: T.Any) -> str:
    if field is None:
        return '_'

    if isinstance(field, dict):
        if field == {}:
            return '_'

        fields = []
        for key, value in field.items():
            if value is None:
                value = "_"
            if value == "":
                fields.append(key)
                continue
            if isinstance(value, SERIALIZABLE_TERMINAL_VALUE_TYPES):
                fields.append(f'{key}={value}')
                continue
            else:
                value_type = type(value)
                raise TypeError(f"Received non-serializable field value of type {value_type}:\n{value}")

        return '|'.join(fields)

    if isinstance(field, tuple):
        return "".join([serialize_field(item) for item in field])

    if isinstance(field, list):
        if len(field[0]) != 2:
            raise ParseException("Can't serialize '{}', invalid format".format(field))
        return "|".join([serialize_field(value) + ":" + str(key) for key, value in field])

    return "{}".format(field)

def serialize(tokenlist: 'TokenList') -> str:
    lines = []

    if tokenlist.metadata:
        for key, value in tokenlist.metadata.items():
            if value:
                line = f"# {key} = {value}"
            else:
                line = f"# {key}"
            lines.append(line)

    for token_data in tokenlist:
        line = '\t'.join(serialize_field(val) for val in token_data.values())
        lines.append(line)

    return '\n'.join(lines) + "\n\n"



import re
import unittest
from io import StringIO
from textwrap import dedent

import pytest

from conllu import parse, parse_incr, parse_tree, parse_tree_incr
from conllu.models import Token, TokenList
from conllu.parser import parse_dict_value, parse_int_value
from tests.helpers import capture_print


@pytest.mark.integration
class TestParse(unittest.TestCase):
    maxDiff = None

    data = re.sub(r"  +", r"\t", dedent("""\
        # text = The quick brown fox jumps over the lazy dog.
        1   The     the    DET    DT   Definite=Def|PronType=Art   4   det     _   _
        2   quick   quick  ADJ    JJ   Degree=Pos                  4   amod    _   _
        3   brown   brown  ADJ    JJ   Degree=Pos                  4   amod    _   _
        4   fox     fox    NOUN   NN   Number=Sing                 5   nsubj   _   _
        5   jumps   jump   VERB   VBZ  Mood=Ind|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin   0   root    _   _
        6   over    over   ADP    IN   _                           9   case    _   _
        7   the     the    DET    DT   Definite=Def|PronType=Art   9   det     _   _
        8   lazy    lazy   ADJ    JJ   Degree=Pos                  9   amod    _   _
        9   dog     dog    NOUN   NN   Number=Sing                 5   nmod    _   SpaceAfter=No
        10  .       .      PUNCT  .    _                           5   punct   _   _

    """))

    def test_parse(self):
        sentences = parse(self.data)
        self.assertEqual(len(sentences), 1)

        sentence = sentences[0]

        self.assertEqual(
            str(sentence),
            "TokenList<The, quick, brown, fox, jumps, over, the, lazy, dog, ."
            ", metadata={text: \"The quick brown fox jumps over the lazy dog.\"}>"
        )

        self.assertEqual(
            sentence[0],
            Token([
                ('id', 1),
                ('form', 'The'),
                ('lemma', 'the'),
                ('upos', 'DET'),
                ('xpos', 'DT'),
                ('feats', Token([('Definite', 'Def'), ('PronType', 'Art')])),
                ('head', 4),
                ('deprel', 'det'),
                ('deps', None),
                ('misc', None)
            ])
        )
        self.assertEqual(
            sentence[8],
            Token([
                ('id', 9),
                ('form', 'dog'),
                ('lemma', 'dog'),
                ('upos', 'NOUN'),
                ('xpos', 'NN'),
                ('feats', Token([('Number', 'Sing')])),
                ('head', 5),
                ('deprel', 'nmod'),
                ('deps', None),
                ('misc', Token([("SpaceAfter", "No")]))
            ])
        )
        self.assertEqual(
            [token["form"] for token in sentence],
            "The quick brown fox jumps over the lazy dog .".split(" ")
        )

        self.assertEqual(
            sentence.metadata["text"],
            "The quick brown fox jumps over the lazy dog."
        )

    def test_parse_tree(self):
        sentences = parse_tree(self.data)
        self.assertEqual(len(sentences), 1)

        root = sentences[0]
        self.assertEqual(str(root), "TokenTree<token={id=5, form=jumps}, children=[...]>")

        self.assertEqual(
            root.token,
            Token([
                ('id', 5),
                ('form', 'jumps'),
                ('lemma', 'jump'),
                ('upos', 'VERB'),
                ('xpos', 'VBZ'),
                ('feats', Token([
                    ("Mood", "Ind"),
                    ("Number", "Sing"),
                    ("Person", "3"),
                    ("Tense", "Pres"),
                    ("VerbForm", "Fin"),
                ])),
                ('head', 0),
                ('deprel', 'root'),
                ('deps', None),
                ('misc', None)
            ])
        )

        self.assertEqual(
            [str(child) for child in root.children],
            [
                "TokenTree<token={id=4, form=fox}, children=[...]>",
                "TokenTree<token={id=9, form=dog}, children=[...]>",
                "TokenTree<token={id=10, form=.}, children=None>",
            ]
        )

        self.assertEqual(
            root.metadata["text"],
            "The quick brown fox jumps over the lazy dog."
        )

        self.assertEqual(root.serialize(), self.data)

        self.assertEqual(
            capture_print(root.print_tree),
            dedent("""\
                (deprel:root) form:jumps lemma:jump upos:VERB [5]
                    (deprel:nsubj) form:fox lemma:fox upos:NOUN [4]
                        (deprel:det) form:The lemma:the upos:DET [1]
                        (deprel:amod) form:quick lemma:quick upos:ADJ [2]
                        (deprel:amod) form:brown lemma:brown upos:ADJ [3]
                    (deprel:nmod) form:dog lemma:dog upos:NOUN [9]
                        (deprel:case) form:over lemma:over upos:ADP [6]
                        (deprel:det) form:the lemma:the upos:DET [7]
                        (deprel:amod) form:lazy lemma:lazy upos:ADJ [8]
                    (deprel:punct) form:. lemma:. upos:PUNCT [10]
            """)
        )

    def test_parse_incr(self):
        self.assertEqual(parse(self.data), list(parse_incr(StringIO(self.data))))

    def test_parse_incr_invalid_file(self):
        with self.assertRaises(FileNotFoundError):
            list(parse_incr("SOME STRING DATA"))

    def test_parse_tree_incr(self):
        self.assertEqual(parse_tree(self.data), list(parse_tree_incr(StringIO(self.data))))


@pytest.mark.integration
class TestTrickyCases(unittest.TestCase):
    maxDiff = None

    def test_parse_and_serialize(self):
        from tests.fixtures import TESTCASES

        for testcase in TESTCASES:
            self.assertEqual(parse(testcase)[0].serialize(), testcase)

    def test_parse_tree_and_serialize(self):
        from tests.fixtures import TESTCASES

        for testcase in TESTCASES:
            data = parse(testcase)
            testcase_without_range_and_elided = TokenList([
                token
                for token in data[0]
                if isinstance(token["id"], int)
            ], metadata=data[0].metadata)
            self.assertEqual(parse_tree(testcase)[0].serialize(), testcase_without_range_and_elided.serialize())


@pytest.mark.integration
class TestParseCoNLLUPlus(unittest.TestCase):
    def test_parse_conllu_plus(self):
        # Note: global.columns affects both sentences
        data = dedent("""\
            # global.columns = ID FORM UPOS HEAD DEPREL MISC PARSEME:MWE
            # source_sent_id = conllu http://hdl.handle.net/11234/1-2837 UD_German-GSD/de_gsd-ud-train.conllu
            # sent_id = train-s16
            # text = Der CDU-Politiker strebt
            1\tDer\tDET\t2\tdet\t_\t*
            2\tCDU\tPROPN\t4\tcompound\tSpaceAfter=No\t*
            3\t-\tPUNCT\t2\tpunct\tSpaceAfter=No\t*
            4\tPolitiker\tNOUN\t5\tnsubj\t_\t*
            5\tstrebt\tVERB\t0\troot\t_\t2:VPC.full

            # source_sent_id = conllu http://hdl.handle.net/11234/1-2837 UD_German-GSD/de_gsd-ud-train.conllu
            # sent_id = train-s17
            # text = Der ortsüblichen Vergleichsmiete orientieren.
            1\tDer\tDET\t19\tdet\t_\t*
            2\tortsüblichen\tADJ\t19\tamod\t_\t*
            3\tVergleichsmiete\tNOUN\t20\tobl\t_\t*
            4\torientieren\tVERB\t8\tacl\tSpaceAfter=No\t1
            5\t.\tPUNCT\t5\tpunct\t_\t*
        """)

        sentences = parse(data)

        self.assertEqual(
            [
                {"form": token["form"], "parseme:mwe": token["parseme:mwe"]}
                for token in sentences[0]
            ],
            [
                {"form": "Der", "parseme:mwe": "*"},
                {"form": "CDU", "parseme:mwe": "*"},
                {"form": "-", "parseme:mwe": "*"},
                {"form": "Politiker", "parseme:mwe": "*"},
                {"form": "strebt", "parseme:mwe": "2:VPC.full"},
            ]
        )
        self.assertEqual(sentences[0].metadata, {
            "global.columns": "ID FORM UPOS HEAD DEPREL MISC PARSEME:MWE",
            "source_sent_id": "conllu http://hdl.handle.net/11234/1-2837 UD_German-GSD/de_gsd-ud-train.conllu",
            "sent_id": "train-s16",
            "text": "Der CDU-Politiker strebt",
        })
        self.assertEqual(
            [
                {"form": token["form"], "parseme:mwe": token["parseme:mwe"]}
                for token in sentences[1]
            ],
            [
                {"form": "Der", "parseme:mwe": "*"},
                {"form": "ortsüblichen", "parseme:mwe": "*"},
                {"form": "Vergleichsmiete", "parseme:mwe": "*"},
                {"form": "orientieren", "parseme:mwe": "1"},
                {"form": ".", "parseme:mwe": "*"},
            ]
        )
        self.assertEqual(sentences[1].metadata, {
            "source_sent_id": "conllu http://hdl.handle.net/11234/1-2837 UD_German-GSD/de_gsd-ud-train.conllu",
            "sent_id": "train-s17",
            "text": "Der ortsüblichen Vergleichsmiete orientieren.",
        })


@pytest.mark.integration
class TestParseCoNLL2009(unittest.TestCase):
    def test_parse_CoNLL2009_1(self):
        data = dedent("""\
            #\tid\tform\tlemma\tplemma\tpos\tppos\tfeats\tpfeats\thead\tphead\tdeprel\tpdeprel\tfillpred\tpred\tapreds
            1\tZ\tz\tz\tR\tR\tSubPOS=R|Cas=2\tSubPOS=R|Cas=2\t10\t10\tAuxP\tAuxP\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_
            2\ttéto\ttento\ttento\tP\tP\tSubPOS=D|Gen=F|Num=S|Cas=2\tSubPOS=D|Gen=F|Num=S|Cas=2\t3\t3\tAtr\tAtr\tY\ttento\t_\tRSTR\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_\t_
            3\tknihy\tkniha\tkniha\tN\tN\tSubPOS=N|Gen=F|Num=S|Cas=2|Neg=A\tSubPOS=N|Gen=F|Num=S|Cas=2|Neg=A\t1\t1\tAdv\tAdv\tY\tkniha\t_\t_\t_\t_\t_\t_\t_\tDIR1\t_\t_\t_\t_\t_\t_\t_\t_

        """)

        sentences = parse(
            data,
            fields=(
                'id', 'form', 'lemma', 'plemma', 'pos', 'ppos', 'feats', 'pfeats',
                'head', 'phead', 'deprel', 'pdeprel', 'fillpred', 'pred', 'apreds'
            ),
            field_parsers={
                "pfeats": lambda line, i: parse_dict_value(line[i]),
                "phead": lambda line, i: parse_int_value(line[i]),
                "apreds": lambda line, i: [
                    apred_field if apred_field != "_" else None
                    for apred_field in line[i:len(line)]
                ],
            },
        )
        self.assertEqual(
            sentences[0][2],
            Token([
                ('id', 3),
                ('form', 'knihy'),
                ('lemma', 'kniha'),
                ('plemma', 'kniha'),
                ('pos', 'N'),
                ('ppos', 'N'),
                ('feats', Token([
                    ('SubPOS', 'N'),
                    ('Gen', 'F'),
                    ('Num', 'S'),
                    ('Cas', '2'),
                    ('Neg', 'A')
                ])),
                ('pfeats', Token([
                    ('SubPOS', 'N'),
                    ('Gen', 'F'),
                    ('Num', 'S'),
                    ('Cas', '2'),
                    ('Neg', 'A')
                ])),
                ('head', 1),
                ('phead', 1),
                ('deprel', 'Adv'),
                ('pdeprel', 'Adv'),
                ('fillpred', 'Y'),
                ('pred', 'kniha'),
                ('apreds', [
                    None, None, None, None, None, None, None, 'DIR1',
                    None, None, None, None, None, None, None, None
                ])
            ])
        )

    def test_parse_CoNLL2009_2(self):
        data = dedent("""\
            #\tid='1'-document_id='36:1047'-span='1'
            1\t+\t+\tPunc\tPunc\t_\t0\tROOT\t_\t_
            2\tIn\tin\tr\tr\tr|-|-|-|-|-|-|-|-\t5\tAuxP\t_\t_
            3\tDei\tDeus\tn\tPropn\tn|-|s|-|-|-|m|g|-\t4\tATR\t_\t_
            4\tnomine\tnomen\tn\tn\tn|-|s|-|-|-|n|b|-\t2\tADV\t_\t_
            5\tregnante\tregno\tt\tt\tt|-|s|p|p|a|m|b|-\t0\tADV\t_\t_

        """)

        sentences = parse(
            data,
            fields=(
                'id', 'form', 'lemma', 'upostag', 'xpostag', 'feats', 'head', 'deprel', 'deps', 'misc'
            ),
            field_parsers={
                "feats": lambda line, i: [feat for feat in line[i].split("|")]
            }
        )
        self.assertEqual(
            sentences[0][4],
            Token([
                ('id', 5),
                ('form', 'regnante'),
                ('lemma', 'regno'),
                ('upostag', 't'),
                ('xpostag', 't'),
                ('feats', ['t', '-', 's', 'p', 'p', 'a', 'm', 'b', '-']),
                ('head', 0),
                ('deprel', 'ADV'),
                ('deps', None),
                ('misc', None),
            ])
        )
        self.assertEqual(
            sentences[0].metadata,
            Token([
                ('id', "'1'-document_id='36:1047'-span='1'")
            ])
        )



import re
import typing as T

from conllu.exceptions import ParseException
from conllu.models import Metadata, Token, TokenList

_IdType = T.Union[int, T.Tuple[int, str, int]]
_FieldParserType = T.Callable[[T.List[str], int], T.Any]
_MetadataParserType = T.Callable[[str, T.Optional[str]], T.Any]

DEFAULT_FIELDS = ('id', 'form', 'lemma', 'upos', 'xpos', 'feats', 'head', 'deprel', 'deps', 'misc')
DEFAULT_FIELD_PARSERS: T.Dict[str, _FieldParserType] = {
    "id": lambda line, i: parse_id_value(line[i]),
    "xpos": lambda line, i: parse_nullable_value(line[i]),
    "feats": lambda line, i: parse_dict_value(line[i]),
    "head": lambda line, i: parse_int_value(line[i]),
    "deps": lambda line, i: parse_paired_list_value(line[i]),
    "misc": lambda line, i: parse_dict_value(line[i]),
}
DEFAULT_METADATA_PARSERS: T.Dict[str, _MetadataParserType] = {
    "newpar": lambda key, value: (key, value),
    "newdoc": lambda key, value: (key, value),
}

def parse_conllu_plus_fields(in_file: T.TextIO,
                             metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
                             ) -> T.Optional[T.Sequence[str]]:
    pos = in_file.tell()

    # Get first line
    try:
        first_sentence = next(parse_sentences(in_file))
        first_line = first_sentence.split("\n")[0]
    except StopIteration:
        first_line = ""

    # parse_sentences moves to file cursor, so reset it here
    in_file.seek(pos)

    if not first_line.startswith("#"):
        return None

    tokenlist = parse_token_and_metadata(first_line, metadata_parsers=metadata_parsers)
    metadata = tokenlist.metadata

    fields = None
    if "global.columns" in metadata and metadata["global.columns"]:
        fields = [value.lower() for value in metadata["global.columns"].split(" ")]

    return fields

def parse_sentences(in_file: T.TextIO) -> T.Iterator[str]:
    buf: T.List[str] = []
    for line in in_file:
        if line.strip() == "":
            if not buf:
                continue
            yield "".join(buf).rstrip()
            buf = []
        else:
            buf.append(line)
    if buf:
        yield "".join(buf).rstrip()

def parse_token_and_metadata(data: str, fields: T.Optional[T.Sequence[str]] = None,
                             field_parsers: T.Optional[T.Dict[str, _FieldParserType]] = None,
                             metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
                             ) -> TokenList:

    if not data:
        raise ParseException("Can't create TokenList, no data sent to constructor.")

    fields = fields or DEFAULT_FIELDS

    if not field_parsers:
        field_parsers = DEFAULT_FIELD_PARSERS.copy()
    elif sorted(field_parsers.keys()) != sorted(fields):
        new_field_parsers = DEFAULT_FIELD_PARSERS.copy()
        new_field_parsers.update(field_parsers)
        field_parsers = new_field_parsers

    tokens = []
    metadata = Metadata()

    for line in data.split('\n'):
        line = line.strip()

        if not line:
            continue

        if line.startswith('#'):
            pairs = parse_comment_line(line, metadata_parsers=metadata_parsers)
            for key, value in pairs:
                metadata[key] = value
        else:
            tokens.append(parse_line(line, fields, field_parsers))

    return TokenList(tokens, metadata, default_fields=fields)

def parse_line(line: str,
               fields: T.Sequence[str], field_parsers: T.Optional[T.Dict[str, _FieldParserType]] = None
               ) -> Token:
    # Be backwards compatible if people called parse_line without field_parsers before
    field_parsers = field_parsers or DEFAULT_FIELD_PARSERS

    # Support xpostag/upostag as aliases for xpos/upos (both ways)
    if "xpostag" not in field_parsers and "xpos" in field_parsers:
        field_parsers["xpostag"] = field_parsers["xpos"]
    if "xpos" not in field_parsers and "xpostag" in field_parsers:
        field_parsers["xpos"] = field_parsers["xpostag"]

    if "upostag" not in field_parsers and "upos" in field_parsers:
        field_parsers["upostag"] = field_parsers["upos"]
    if "upos" not in field_parsers and "upostag" in field_parsers:
        field_parsers["upos"] = field_parsers["upostag"]

    line_split = re.split(r"\t| {2,}", line)

    if len(line_split) == 1:
        raise ParseException("Invalid line format, line must contain either tabs or two spaces.")

    data = Token()

    for i, field in enumerate(fields):
        # Allow parsing CoNNL-U files with fewer columns
        if i >= len(line_split):
            break

        if field in field_parsers:
            try:
                value = field_parsers[field](line_split, i)
            except ParseException as e:
                raise ParseException("Failed parsing field '{}': ".format(field) + str(e))

        else:
            value = line_split[i]

        data[str(field)] = value

    return data

def parse_comment_line(line: str,
                       metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
                       ) -> T.List[T.Tuple[str, T.Optional[str]]]:
    line = line.strip()

    if line[0] != '#':
        raise ParseException("Invalid comment format, comment must start with '#'")

    key, value = parse_pair_value(line[1:])

    if not metadata_parsers:
        metadata_parsers = DEFAULT_METADATA_PARSERS.copy()
    else:
        new_metadata_parsers = DEFAULT_METADATA_PARSERS.copy()
        new_metadata_parsers.update(metadata_parsers)
        metadata_parsers = new_metadata_parsers

    custom_result = None
    if key in metadata_parsers:
        custom_result = metadata_parsers[key](key, value)
    elif "__fallback__" in metadata_parsers:
        custom_result = metadata_parsers["__fallback__"](key, value)

    # Allow returning pair instead of list of pairs from metadata parsers
    if custom_result:
        if isinstance(custom_result, tuple):
            key, value = custom_result
            return [(str(key), value)]
        return [(str(key), value) for key, value in custom_result]

    if not key or not value:
        # Lines without value are invalid by default
        return []

    return [(str(key), value)]

def parse_pair_value(value: str) -> T.Tuple[str, T.Optional[str]]:
    key_maybe_value = value.split('=', 1)
    key = key_maybe_value[0].strip()
    value_new = None if len(key_maybe_value) == 1 else key_maybe_value[1].strip()

    return key, value_new


INTEGER = re.compile(r"0|(\-?[1-9][0-9]*)")

def parse_int_value(value: str) -> T.Optional[int]:
    if value == '_':
        return None

    if re.fullmatch(INTEGER, value):
        return int(value)
    else:
        raise ParseException("'{}' is not a valid value for parse_int_value.".format(value))


ID_SINGLE = re.compile(r"(?:0|[1-9][0-9]*)")
ID_RANGE = re.compile(r"[1-9][0-9]*\-[1-9][0-9]*")
ID_DOT_ID = re.compile(r"[0-9][0-9]*\.[1-9][0-9]*")

def parse_id_value(value: str) -> T.Optional[_IdType]:
    if not value or value == '_':
        return None

    if re.fullmatch(ID_SINGLE, value):
        return int(value)

    elif re.fullmatch(ID_RANGE, value):
        from_str, to_str = value.split("-")
        from_, to = int(from_str), int(to_str)
        if to >= from_:
            return (int(from_), "-", int(to))

    elif re.fullmatch(ID_DOT_ID, value):
        return (int(value.split(".")[0]), ".", int(value.split(".")[1]))

    raise ParseException("'{}' is not a valid ID.".format(value))


ANY_ID = re.compile(ID_SINGLE.pattern + "|" + ID_RANGE.pattern + "|" + ID_DOT_ID.pattern)
DEPS_RE = re.compile("(" + ANY_ID.pattern + r")(:[^\d:_\-|][^:|]*)+")
MULTI_DEPS_PATTERN = re.compile(r"{}(\|{})*".format(DEPS_RE.pattern, DEPS_RE.pattern))

def parse_paired_list_value(value: str) -> T.Union[T.Optional[str], T.List[T.Tuple[str, T.Optional[_IdType]]]]:
    if re.fullmatch(MULTI_DEPS_PATTERN, value):
        return [
            (part.split(":", 1)[1], parse_id_value(part.split(":")[0]))
            for part in value.split("|")
        ]

    return parse_nullable_value(value)

def parse_dict_value(value: str) -> T.Optional[T.Dict[str, T.Optional[str]]]:
    if parse_nullable_value(value) is None:
        return None

    return {
        part.split("=")[0]: parse_nullable_value(part.split("=")[1]) if "=" in part else ""
        for part in value.split("|") if parse_nullable_value(part.split("=")[0]) is not None
    }

def parse_nullable_value(value: str) -> T.Optional[str]:
    if not value or value == "_":
        return None

    return value

# DEPRECATED: Mantain old paths until next major version
def serialize(tokenlist: TokenList) -> str:
    from conllu.serializer import serialize as new_serialize
    return new_serialize(tokenlist)

def serialize_field(field: T.Any) -> str:
    from conllu.serializer import serialize_field as new_serialize_field
    return new_serialize_field(field)

def head_to_token(sentence: TokenList) -> T.Dict[int, T.List[Token]]:
    return TokenList.head_to_token(sentence)



import unittest
from io import StringIO
from textwrap import dedent

from conllu.models import Token, TokenList
from conllu.parser import (
    DEFAULT_FIELDS, ParseException, head_to_token, parse_comment_line, parse_conllu_plus_fields, parse_dict_value,
    parse_id_value, parse_int_value, parse_line, parse_nullable_value, parse_paired_list_value, parse_sentences,
    parse_token_and_metadata, serialize, serialize_field,
)


class TestParseConlluPlusFields(unittest.TestCase):
    def test_empty(self):
        self.assertEqual(parse_conllu_plus_fields(StringIO("")), None)
        self.assertEqual(parse_conllu_plus_fields(StringIO(None)), None)

    def test_simple(self):
        data = dedent("""\
            # global.columns = ID FORM UPOS HEAD DEPREL MISC PARSEME:MWE
            1\tDer\tDET\t2\tdet\t_\t*
        """)
        self.assertEqual(
            parse_conllu_plus_fields(StringIO(data)),
            ["id", "form", "upos", "head", "deprel", "misc", "parseme:mwe"]
        )

    def test_empty_columns(self):
        data = dedent("""\
            # global.columns =
            1\tDer\tDET\t2\tdet\t_\t*
        """)
        self.assertEqual(parse_conllu_plus_fields(StringIO(data)), None)

class TestParseSentencesGenerator(unittest.TestCase):
    def test_empty(self):
        self.assertEqual(list(parse_sentences(StringIO(""))), [])
        self.assertEqual(list(parse_sentences(StringIO(None))), [])

    def test_simple(self):
        data = dedent("""\
            1\thej
            2\tdå
            3\thej

            1\thej
            2\tdå
            3\thej
        """)
        sentences = list(parse_sentences(StringIO(data)))
        self.assertEqual(sentences, [
            '1\thej\n2\tdå\n3\thej',
            '1\thej\n2\tdå\n3\thej',
        ])

    def test_multiple_newlines(self):
        data = dedent("""\
            1\thej
            2\tdå


            1\thej
            2\tdå



            1\thej
            2\tdå
        """)
        sentences = list(parse_sentences(StringIO(data)))
        self.assertEqual(sentences, [
            '1\thej\n2\tdå',
            '1\thej\n2\tdå',
            '1\thej\n2\tdå',
        ])

    def test_ends_without_newline(self):
        data = "1\thej\n2\tdå"
        sentences = list(parse_sentences(StringIO(data)))
        self.assertEqual(sentences, [
            '1\thej\n2\tdå',
        ])

    # WNUT 2017 (https://noisy-text.github.io/2017/emerging-rare-entities.html) has blank lines
    # denoted as \t\n, so test this
    def test_newlines_with_tabs(self):
        data = dedent("""\
            1\thej
            2\tdå
            \t
            1\thej
            2\tdå
        """)
        sentences = list(parse_sentences(StringIO(data)))
        self.assertEqual(sentences, [
            '1\thej\n2\tdå',
            '1\thej\n2\tdå',
        ])

class TestParseTokenAndMetadata(unittest.TestCase):
    def test_empty(self):
        with self.assertRaises(ParseException):
            parse_token_and_metadata(None)

    def test_newlines_in_sentence(self):
        data = dedent("""\
            # meta = data
            1\thej
            2\tdå

            3\thej
            4\tdå
        """)
        tokenlist = parse_token_and_metadata(data)
        self.assertListEqual(tokenlist, TokenList([
            Token([("id", 1), ("form", "hej")]),
            Token([("id", 2), ("form", "då")]),
            Token([("id", 3), ("form", "hej")]),
            Token([("id", 4), ("form", "då")]),
        ], metadata={"meta": "data"}))

    def test_invalid_metadata(self):
        data = dedent("""\
            # meta = data2
            # meta = data
            # newdoc
            # newpar
            # meta
            # = data
        """)
        tokenlist = parse_token_and_metadata(data)
        self.assertEqual(tokenlist.metadata, {
            "meta": "data",
            "newdoc": None,
            "newpar": None,
        })

    def test_custom_metadata_parsers(self):
        data = dedent("""\
            # global.columns = ID FORM LEMMA UPOS XPOS FEATS HEAD DEPREL DEPS MISC
            # newdoc id = mf920901-001
            # newpar id = mf920901-001-p1
            # sent_id = mf920901-001-p1s1A
            # text = Slovenská ústava: pro i proti
            # text_en = Slovak constitution: pros and cons
        """)
        tokenlist = parse_token_and_metadata(data)
        self.assertEqual(tokenlist.metadata, Token([
            ("global.columns", "ID FORM LEMMA UPOS XPOS FEATS HEAD DEPREL DEPS MISC"),
            ("newdoc id", "mf920901-001"),
            ("newpar id", "mf920901-001-p1"),
            ("sent_id", "mf920901-001-p1s1A"),
            ("text", "Slovenská ústava: pro i proti"),
            ("text_en", "Slovak constitution: pros and cons"),
        ]))

        tokenlist = parse_token_and_metadata(
            data,
            metadata_parsers={"global.columns": lambda key, value: (key, value.split())}
        )
        self.assertEqual(tokenlist.metadata, Token([
            ("global.columns", ["ID", "FORM", "LEMMA", "UPOS", "XPOS", "FEATS", "HEAD", "DEPREL", "DEPS", "MISC"]),
            ("newdoc id", "mf920901-001"),
            ("newpar id", "mf920901-001-p1"),
            ("sent_id", "mf920901-001-p1s1A"),
            ("text", "Slovenská ústava: pro i proti"),
            ("text_en", "Slovak constitution: pros and cons"),
        ]))

    def test_one_to_many_custom_metadata_parser(self):
        data = dedent("""\
            #\tid='1'-document_id='36:1047'-span='1'
        """)

        tokenlist = parse_token_and_metadata(
            data,
            metadata_parsers={
                "id": lambda key, value: [
                    (pair.split("=", 1)[0], pair.split("=", 1)[1].strip("'"))
                    for pair in (key + "=" + value).split("-")
                ]
            },
        )
        self.assertEqual(tokenlist.metadata, Token([
            ("id", "1"),
            ("document_id", "36:1047"),
            ("span", "1"),
        ]))

    def test_fallback_metadata_parser(self):
        data = dedent("""\
            #20191005
        """)

        tokenlist = parse_token_and_metadata(
            data,
            metadata_parsers={
                "__fallback__": lambda key, value: ("sentence-id", key),
            },
        )
        self.assertEqual(tokenlist.metadata, Token([
            ("sentence-id", "20191005"),
        ]))

    def test_custom_fields(self):
        data = dedent("""\
            1\t1\t1
            2\t2\t2
        """)
        tokenlist = parse_token_and_metadata(data, fields=("id", "id", "id"))
        self.assertEqual(tokenlist, [
            Token([("id", 1), ("id", 1), ("id", 1)]),
            Token([("id", 2), ("id", 2), ("id", 2)]),
        ])

    def test_custom_field_parsers(self):
        data = dedent("""\
            1\tbackwards\tline
            2\tparis\tsirap
        """)
        fields = ("id", "backwards")

        # A field parser that takes all remaining field, reverses their letters and joins them
        def parse_backwards(value):
            return " ".join([part[::-1] for part in value])

        # This overrides the default parsers, so the id is parsed as a string
        field_parsers = {
            "id": lambda line, i: line[i],
            "backwards": lambda line, i: parse_backwards(line[i:len(line)])
        }

        tokenlist = parse_token_and_metadata(data, fields=fields, field_parsers=field_parsers)
        self.assertEqual(tokenlist, [
            Token([("id", '1'), ("backwards", "sdrawkcab enil")]),
            Token([("id", '2'), ("backwards", "sirap paris")]),
        ])

    def test_default_field_parsers_when_undefined(self):
        data = dedent("""\
            1\tfrom
            2\tparis
        """)
        fields = ("id", "form")
        field_parsers = {
            # Rely on default 'id' field parser
            "form": lambda line, i: line[i].upper()
        }
        tokenlist = parse_token_and_metadata(data, fields=fields, field_parsers=field_parsers)
        self.assertEqual(tokenlist, [
            Token([("id", 1), ("form", "FROM")]),
            Token([("id", 2), ("form", "PARIS")]),
        ])

    def test_fields_from_parse_become_defaults_in_tokenlist(self):
        data = dedent("""\
            1\thej
            2\tdå
        """)
        tokenlist = parse_token_and_metadata(data, fields=["id", "form"])
        tokenlist.append({"id": 3})
        self.assertEqual(tokenlist, [
            Token({"id": 1, "form": "hej"}),
            Token({"id": 2, "form": "då"}),
            Token({"id": 3, "form": "_"}),  # Form is set to default value
        ])

class TestParseLine(unittest.TestCase):
    def test_empty(self):
        with self.assertRaises(ParseException) as assert_context:
            line = "invalid_id\t_\t_\t_\t_\t_\t_\t_\t_\t"
            parse_line(line, fields=DEFAULT_FIELDS)

        expected = "Failed parsing field 'id'"
        self.assertEqual(str(assert_context.exception)[:len(expected)], expected)

    def test_parse_line(self):
        line = "1\tThe\tthe\tDET\tDT\tDefinite=Def|PronType=Art\t4\tdet\t_\t_"
        self.assertEqual(
            parse_line(line, fields=DEFAULT_FIELDS),
            Token([
                ('id', 1),
                ('form', 'The'),
                ('lemma', 'the'),
                ('upos', 'DET'),
                ('xpos', 'DT'),
                ('feats', Token([('Definite', 'Def'), ('PronType', 'Art')])),
                ('head', 4),
                ('deprel', 'det'),
                ('deps', None),
                ('misc', None)
            ])
        )

    def test_parse_line_nullable_fields(self):
        line = "_\t_\t_\t_\t_\t_\t_\t_\t_\t_"
        self.assertEqual(
            parse_line(line, fields=DEFAULT_FIELDS),
            Token([
                ('id', None),
                ('form', '_'),
                ('lemma', '_'),
                ('upos', '_'),
                ('xpos', None),
                ('feats', None),
                ('head', None),
                ('deprel', '_'),
                ('deps', None),
                ('misc', None)
            ])
        )

    def test_parse_line_only_id_head(self):
        line = "1\tThe\tthe\tDET\tDT\tDefinite=Def|PronType=Art\t4\tdet\t_\t_"
        self.assertEqual(parse_line(line, fields=["id", "form"]), Token([
            ('id', 1),
            ('form', 'The'),
        ]))

    def test_parse_line_fewer_columns(self):
        line = "1\tThe\tthe\tDET\tDT"
        self.assertEqual(parse_line(line, fields=DEFAULT_FIELDS), Token([
            ('id', 1),
            ('form', 'The'),
            ('lemma', 'the'),
            ('upos', 'DET'),
            ('xpos', 'DT'),
        ]))

    def test_parse_line_without_spacing(self):
        line = "1ThetheDETDTDefinite=Def|PronType=Art4det__"
        with self.assertRaises(ParseException) as assert_context:
            parse_line(line, fields=DEFAULT_FIELDS)

        expected = "Invalid line format"
        self.assertEqual(str(assert_context.exception)[:len(expected)], expected)

    def test_parse_line_with_spaces(self):
        line = "1 The the DET DT Definite=Def|PronType=Art 4 det _ _"
        with self.assertRaises(ParseException) as assert_context:
            parse_line(line, fields=DEFAULT_FIELDS)

        expected = "Invalid line format"
        self.assertEqual(str(assert_context.exception)[:len(expected)], expected)

    def test_parse_line_two_spaces(self):
        line = "1  The  the  DET  DT  Definite=Def|PronType=Art  4  det  _  _"
        self.assertEqual(parse_line(line, fields=["id", "form"]), Token([
            ('id', 1),
            ('form', 'The'),
        ]))

    def test_parse_custom_fieldparsers(self):
        line = "1\t2"
        custom_fieldparsers = {
            "id": lambda line, i: line[i] * 5,
        }
        self.assertEqual(
            parse_line(line, fields=["id"], field_parsers=custom_fieldparsers),
            Token([
                ('id', "11111"),
            ])
        )

    def test_parse_fieldparsers_alias_xupostag(self):
        line = "1\t2"
        custom_fieldparsers = {
            "xpostag": lambda line, i: line[i] * 5,
            "upostag": lambda line, i: line[i] * 5,
        }
        self.assertEqual(
            parse_line(line, fields=["xpos", "upos"], field_parsers=custom_fieldparsers),
            Token([
                ('xpos', "11111"),
                ('upos', "22222"),
            ])
        )

    def test_parse_fieldparsers_alias_xupos(self):
        line = "1\t2"
        custom_fieldparsers = {
            "xpos": lambda line, i: line[i] * 5,
            "upos": lambda line, i: line[i] * 5,
        }
        self.assertEqual(
            parse_line(line, fields=["xpostag", "upostag"], field_parsers=custom_fieldparsers),
            Token([
                ('xpostag', "11111"),
                ('upostag', "22222"),
            ])
        )

    def test_parse_fieldparsers_doesnt_alias_when_exists(self):
        line = "1\t2"
        custom_fieldparsers = {
            "xpos": lambda line, i: line[i] * 5,
            "xpostag": lambda line, i: line[i],
            "upos": lambda line, i: line[i] * 5,
            "upostag": lambda line, i: line[i],
        }
        self.assertEqual(
            parse_line(line, fields=["xpostag", "upostag"], field_parsers=custom_fieldparsers),
            Token([
                ('xpostag', "1"),
                ('upostag', "2"),
            ])
        )

    def test_parse_fieldparsers_alias_two_ways(self):
        line = "1\t2"
        custom_fieldparsers = {
            "xpos": lambda line, i: line[i] * 5,
            "upostag": lambda line, i: line[i] * 5,
        }
        self.assertEqual(
            parse_line(line, fields=["xpostag", "upos"], field_parsers=custom_fieldparsers),
            Token([
                ('xpostag', "11111"),
                ('upos', "22222"),
            ])
        )

class TestParseCommentLine(unittest.TestCase):
    def test_parse_spaces_before_square(self):
        data = ["# a = 1", "  # a = 1", "\t# a = 1"]
        for item in data:
            self.assertEqual(parse_comment_line(item), [("a", "1")])

    def test_parse_comment_line(self):
        data = "# sent_id = 1"
        self.assertEqual(parse_comment_line(data), [("sent_id", "1")])

    def test_parse_comment_line_multiple_equals(self):
        data = "# text = five plus three = eight"
        self.assertEqual(parse_comment_line(data), [("text", "five plus three = eight")])

    def test_parse_comment_line_without_square(self):
        data = "sent_id = 1"
        with self.assertRaises(ParseException):
            parse_comment_line(data)

    def test_parse_comment_line_without_equals(self):
        data = "# sent_id: 1"
        self.assertEqual(parse_comment_line(data), [])

    def test_parse_comment_line_without_space(self):
        data = "#sent_id = 1"
        self.assertEqual(parse_comment_line(data), [("sent_id", "1")])

    def test_parse_comment_line_optional_value(self):
        data = '# newdoc'
        self.assertEqual(parse_comment_line(data), [("newdoc", None)])
        data = '# newpar'
        self.assertEqual(parse_comment_line(data), [("newpar", None)])
        data = '# invalid'
        self.assertEqual(parse_comment_line(data), [])

    def test_parse_comment_line_optional_value_no_space(self):
        data = '#newdoc'
        self.assertEqual(parse_comment_line(data), [("newdoc", None)])
        data = '#newpar'
        self.assertEqual(parse_comment_line(data), [("newpar", None)])
        data = '#invalid'
        self.assertEqual(parse_comment_line(data), [])


class TestParseIntValue(unittest.TestCase):
    def test_parse_int_value(self):
        self.assertEqual(parse_int_value("_"), None)
        self.assertEqual(parse_int_value("4"), 4)
        self.assertEqual(parse_int_value("0"), 0)
        self.assertEqual(parse_int_value("10"), 10)
        self.assertEqual(parse_int_value("-10"), -10)

        with self.assertRaises(ParseException):
            parse_int_value("a")

class TestParseIDValue(unittest.TestCase):
    def _run_valid_invalid_tests(self, valid, invalid=[]):
        for value, result in valid:
            self.assertEqual(parse_id_value(value), result)

        for value in invalid:
            with self.assertRaises(ParseException):
                parse_id_value(value)

    def test_none(self):
        self._run_valid_invalid_tests([
            ("_", None),
            (None, None),
        ])

    def test_single(self):
        self._run_valid_invalid_tests([
            ("4", 4),
            ("10", 10),
            ("0", 0),
        ], [
            "-10",
            "a",
        ])

    def test_range(self):
        self._run_valid_invalid_tests([
            ("1-2", (1, "-", 2)),
            ("1-999", (1, "-", 999)),
            ("1000-1001", (1000, "-", 1001)),
            ("1-1", (1, "-", 1)),
        ], [
            "2-1",
            "1-0",
            "0-1-2",
            "0-a",
            "a-0",
        ])

    def test_decimal(self):
        self._run_valid_invalid_tests([
            ("0.1", (0, ".", 1)),
            ("1.1", (1, ".", 1)),
            ("1.10", (1, ".", 10)),
            ("10.1", (10, ".", 1)),
        ], [
            "1.0",
            "a.0",
            "0.a",
        ])

class TestParsePairedListValue(unittest.TestCase):
    def test_parse_paired_list(self):
        self.assertEqual(
            parse_paired_list_value("4:nsubj"),
            [("nsubj", 4)]
        )
        self.assertEqual(
            parse_paired_list_value("1:under_case_:under_case_|2:dash-sign-:dash-sign-"),
            [("under_case_:under_case_", 1), ("dash-sign-:dash-sign-", 2)]
        )
        self.assertEqual(
            parse_paired_list_value("2:nsubj|4:nsubj"),
            [("nsubj", 2), ("nsubj", 4)]
        )
        self.assertEqual(
            parse_paired_list_value("1:flat:name|2:amod|20:nsubj"),
            [("flat:name", 1), ("amod", 2), ("nsubj", 20)]
        )
        self.assertEqual(
            parse_paired_list_value("5:conj:and|8.1:nsubj:pass|9:nsubj:xsubj"),
            [("conj:and", 5), ("nsubj:pass", (8, ".", 1)), ("nsubj:xsubj", 9)]
        )
        self.assertEqual(
            parse_paired_list_value("1:compound|6:ARG|9:ARG1|10:ARG2"),
            [('compound', 1), ('ARG', 6), ('ARG1', 9), ('ARG2', 10)]
        )
        self.assertNotEqual(
            parse_paired_list_value("1:compound|6:ARG|9:ARG1|10:2ARG"),
            [('compound', 1), ('ARG', 6), ('ARG1', 9), ('2ARG', 10)]
        )
        self.assertEqual(
            parse_paired_list_value("0:root"),
            [('root', 0)]
        )
        self.assertEqual(
            parse_paired_list_value("1:obl:arg:gen"),
            [('obl:arg:gen', 1)]
        )
        self.assertEqual(
            parse_paired_list_value("25:obl:arg:į"),
            [('obl:arg:į', 25)]
        )
        self.assertEqual(
            parse_paired_list_value("3:obl:arg:عَلَى:gen"),
            [('obl:arg:عَلَى:gen', 3)]
        )

    def test_parse_empty(self):
        testcases = [
            "",
            "_",
        ]
        for testcase in testcases:
            # Empty strings should return None
            self.assertEqual(parse_paired_list_value(testcase), None)

    def test_parse_invalid(self):
        testcases = [
            "x:nsubj",
            "0:",
            "0:_",
            "0:-",
            ":",
            ":nsubj",
            "1:nsubj|",
            "1:nsubj||1:nsubj",
        ]
        for testcase in testcases:
            # Invalid strings should be returned untouched
            self.assertEqual(parse_paired_list_value(testcase), testcase)

class TestParseDictValue(unittest.TestCase):
    def test_parse_dict_value(self):
        self.assertEqual(
            parse_dict_value("key1"),
            Token([("key1", "")])
        )
        self.assertEqual(
            parse_dict_value("key1=val1"),
            Token([("key1", "val1")])
        )
        self.assertEqual(
            parse_dict_value("key1=val1|key2=val2"),
            Token([("key1", "val1"), ("key2", "val2")])
        )
        self.assertEqual(
            parse_dict_value("key1=val1|key2|key3=val3"),
            Token([("key1", "val1"), ("key2", ""), ("key3", "val3")])
        )
        self.assertEqual(
            parse_dict_value("key1=val1|key1=val2"),
            Token([("key1", "val2")])
        )
        self.assertEqual(
            parse_dict_value("key1=_|_|_=val1"),
            Token([("key1", None)])
        )
        self.assertEqual(parse_dict_value(""), None)
        self.assertEqual(parse_dict_value("_"), None)

class TestParseNullableValue(unittest.TestCase):
    def test_parse_nullable_value(self):
        self.assertEqual(parse_nullable_value("_"), None)
        self.assertEqual(parse_nullable_value(""), None)
        self.assertEqual(parse_nullable_value("hello"), "hello")

class TestHeadToToken(unittest.TestCase):
    maxDiff = None

    def test_simple(self):
        self.assertEqual(
            head_to_token([
                {"data": "a", "head": 0},
                {"data": "dog", "head": 1},
                {"data": "wags", "head": 2},
                {"data": "tail", "head": 1},
            ]),
            {
                0: [{"data": "a", "head": 0}],
                1: [{"data": "dog", "head": 1}, {"data": "tail", "head": 1}],
                2: [{"data": "wags", "head": 2}],
            }
        )

    def test_missing_head(self):
        with self.assertRaises(ParseException):
            head_to_token([])

        with self.assertRaises(ParseException):
            head_to_token([{"data": "a"}])

    def test_negative_head(self):
        self.assertEqual(
            head_to_token([
                {"data": "a", "head": 0},
                {"data": "dog", "head": 1},
                {"data": "wags", "head": -1},
                {"data": "tail", "head": -2},
            ]),
            {
                0: [{"data": "a", "head": 0}],
                1: [{"data": "dog", "head": 1}],
            }
        )

    def test_none_head(self):
        self.assertEqual(
            head_to_token([
                {"data": "a", "head": 0},
                {"data": "dog", "head": 1},
                {"data": "wags", "head": None},
                {"data": "tail", "head": None},
            ]),
            {
                0: [{"data": "a", "head": 0}],
                1: [{"data": "dog", "head": 1}],
            }
        )

    def test_range_ids_ignored(self):
        self.assertEqual(
            dict(head_to_token([
                {"data": "a", "head": 0},
                {"data": "dog", "head": 1},
                {"data": "a dog", "head": 0, "id": (1, "-", 2)},  # Range node
                {"data": "wags", "head": 2},
                {"data": "tail", "head": 1},
            ])),
            {
                0: [{"data": "a", "head": 0}],
                1: [{"data": "dog", "head": 1}, {"data": "tail", "head": 1}],
                2: [{"data": "wags", "head": 2}],
            }
        )

    def test_decimal_ids_ignored(self):
        self.assertEqual(
            dict(head_to_token([
                {"data": "a", "head": 0},
                {"data": "dog", "head": 1},
                {"data": "that", "head": 1, "id": (1, ".", 1)},
                {"data": "wags", "head": 2},
                {"data": "its", "head": 2, "id": (2, ".", 1)},  # Empty node
                {"data": "tail", "head": 1},
            ])),
            {
                0: [{"data": "a", "head": 0}],
                1: [{"data": "dog", "head": 1}, {"data": "tail", "head": 1}],
                2: [{"data": "wags", "head": 2}],
            }
        )


class TestSerializeField(unittest.TestCase):
    def test_ordered_dict(self):
        data = Token()
        self.assertEqual(serialize_field(data), "_")

        data = Token([('SpaceAfter', 'No')])
        self.assertEqual(serialize_field(data), "SpaceAfter=No")

        data = Token([('Translit', None)])
        self.assertEqual(serialize_field(data), "Translit=_")

    def test_none(self):
        data = None
        self.assertEqual(serialize_field(data), "_")

    def test_string(self):
        data = "ADJ"
        self.assertEqual(serialize_field(data), "ADJ")

    def test_tuple(self):
        data = (1, "-", 2)
        self.assertEqual(serialize_field(data), "1-2")

        data = (100, ".", 2)
        self.assertEqual(serialize_field(data), "100.2")

        with self.assertRaises(ParseException):
            serialize_field([(1, "-", 2)])

    def test_list(self):
        data = [("nsubj", 2), ("nmod", 1)]
        self.assertEqual(serialize_field(data), "2:nsubj|1:nmod")

    def test_blank_dict(self):
        data = {}
        self.assertEqual(serialize_field(data), "_")

    def test_dict_with_blank_value(self):
        data = {"key1": "value1", "key2": ""}
        self.assertEqual(serialize_field(data), "key1=value1|key2")

    def test_dict_with_int_value(self):
        data = {"key1": -1, "key2": 2}
        self.assertEqual(serialize_field(data), "key1=-1|key2=2")

    def test_dict_with_float_value(self):
        data = {"key1": 0.33, "key2": 3.142}
        self.assertEqual(serialize_field(data), "key1=0.33|key2=3.142")

    def test_dict_with_str_value(self):
        data = {"key1": "foo", "key2": "bar"}
        self.assertEqual(serialize_field(data), "key1=foo|key2=bar")

    def test_dict_with_bool_value(self):
        data = {"key1": True, "key2": False}
        self.assertEqual(serialize_field(data), "key1=True|key2=False")

    def test_dict_with_nonserializable_type0(self):
        _object = object()
        data = {"key1": _object}
        with self.assertRaises(TypeError):
            serialize_field(data)


class TestSerialize(unittest.TestCase):
    def test_identity_unicode(self):
        data = "5\tlängtar\n\n"
        tokenlist = parse_token_and_metadata(data)
        self.assertEqual(serialize(tokenlist), data)

    def test_metadata(self):
        data = dedent("""\
            # newdoc
            # data = meta
            # meta = data
            1\tdog

        """)
        tokenlist = parse_token_and_metadata(data)
        self.assertEqual(serialize(tokenlist), data)

    def test_non_str_metadata(self):
        data = dedent("""\
            # line_id = 128
            # weight = 3.14
            1\tdog

        """)
        tokenlist = parse_token_and_metadata(
            data,
            metadata_parsers={
                "line_id": lambda k, v: (k, int(v)),
                "weight": lambda k, v: (k, float(v))
            }
        )
        self.assertIsInstance(tokenlist.metadata['line_id'], int)
        self.assertEqual(tokenlist.metadata['line_id'], 128)

        self.assertIsInstance(tokenlist.metadata['weight'], float)
        self.assertEqual(tokenlist.metadata['weight'], 3.14)

        self.assertEqual(serialize(tokenlist), data)

    def test_serialize_tricky_fields(self):
        data = dedent("""\
            5\tjumps\tjump\tVERB\tVBZ\tMood=Ind|Number=Sing\t0\troot\t_\tSpaceAfter=No
        """)
        tokenlist = parse_token_and_metadata(data)
        self.assertEqual(serialize(tokenlist).strip(), data.strip())



import unittest
from textwrap import dedent

from conllu.models import Metadata, SentenceGenerator, SentenceList, Token, TokenList, TokenTree
from conllu.parser import ParseException, serialize
from tests.helpers import capture_print


class TestToken(unittest.TestCase):
    def test_xupos_to_xupostag(self):
        token = Token({"id": 1, "xpos": "DT", "upos": "DET"})
        self.assertEqual(token["xpos"], "DT")
        self.assertEqual(token["xpostag"], "DT")
        self.assertEqual(token["upos"], "DET")
        self.assertEqual(token["upostag"], "DET")
        self.assertEqual(token.get("xpos"), "DT")
        self.assertEqual(token.get("xpostag"), "DT")
        self.assertEqual(token.get("upos"), "DET")
        self.assertEqual(token.get("upostag"), "DET")

    def test_xupostag_to_xupos(self):
        token = Token({"id": 1, "xpostag": "DT", "upostag": "DET"})
        self.assertEqual(token["xpos"], "DT")
        self.assertEqual(token["xpostag"], "DT")
        self.assertEqual(token["upos"], "DET")
        self.assertEqual(token["upostag"], "DET")
        self.assertEqual(token.get("xpos"), "DT")
        self.assertEqual(token.get("xpostag"), "DT")
        self.assertEqual(token.get("upos"), "DET")
        self.assertEqual(token.get("upostag"), "DET")

    def test_invalid_key_access(self):
        token = Token({"id": 1, "xpostag": "DT", "upostag": "DET"})
        with self.assertRaises(KeyError):
            token["inexistent_value"]

        self.assertEqual(token.get("inexistent_value"), None)
        self.assertEqual(token.get("inexistent_value", "HEJ"), "HEJ")

    def test_missing_upos_and_upostag(self):
        token = Token({"id": 1})
        self.assertEqual(token["upos"], None)
        self.assertEqual(token.get("upos"), None)

    def test_str(self):
        self.assertEqual(str(Token({"id": 1, "form": "The"})), "The")
        self.assertEqual(str(Token({"id": 1})), "id=1")
        self.assertEqual(str(Token({"x": 1})), "")

class TestTokenList(unittest.TestCase):
    def test_constructor(self):
        with self.assertRaises(ParseException):
            TokenList({"id": 1})

    def test_eq(self):
        self.assertEqual(
            TokenList([{"id": 1}]),
            TokenList([{"id": 1}])
        )
        self.assertNotEqual(
            TokenList([{"id": 1}], metadata={"meta": "data"}),
            TokenList([{"id": 1}])
        )
        self.assertEqual(
            TokenList([{"id": 1}], metadata={"meta": "data"}),
            TokenList([{"id": 1}], metadata={"meta": "data"})
        )
        self.assertEqual(
            TokenList([{"id": 1}]),
            [{"id": 1}]
        )

    def test_len(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}, {"id": 3}])
        self.assertEqual(3, len(tokenlist))

    def test_clear(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}, {"id": 3}], {"meta": "data"})
        tokenlist.clear()
        self.assertEqual(len(tokenlist), 0)
        self.assertEqual(tokenlist.metadata, Metadata())

    def test_copy(self):
        tokenlist1 = TokenList([{"id": 1}, {"id": 2}, {"id": 3}], {"meta": "data"})
        tokenlist2 = tokenlist1.copy()
        self.assertIsNot(tokenlist1, tokenlist2)
        self.assertEqual(tokenlist1, tokenlist2)

    def test_extend_tokenlist_no_metadata_with_list(self):
        tokenlist1 = TokenList([{"id": 1}, {"id": 2}, {"id": 3}])
        tokenlist2 = [{"id": 4}, {"id": 5}, {"id": 6}]
        tokenlist1.extend(tokenlist2)
        tokenlist3 = TokenList([{"id": 1}, {"id": 2}, {"id": 3}, {"id": 4}, {"id": 5}, {"id": 6}])
        self.assertEqual(tokenlist1, tokenlist3)

    def test_extend_tokenlist_and_merge_metadata(self):
        tokenlist4 = TokenList([{"id": 1}, {"id": 2}, {"id": 3}], {"meta1": "data1"})
        tokenlist5 = TokenList([{"id": 4}, {"id": 5}, {"id": 6}], {"meta2": "data2"})
        tokenlist4.extend(tokenlist5)
        tokenlist6 = TokenList([{"id": 1}, {"id": 2}, {"id": 3}, {"id": 4}, {"id": 5}, {"id": 6}],
                               {"meta1": "data1", "meta2": "data2"})
        self.assertEqual(tokenlist4, tokenlist6)

    def test_extend_with_dict_list(self):
        tokenlist = TokenList([{"id": 1}])
        tokenlist.extend([{"id": 2}, {"id": 3}])
        self.assertEqual(tokenlist, TokenList([{"id": 1}, {"id": 2}, {"id": 3}]))

    def test_empty(self):
        tokenlist = TokenList()
        self.assertEqual(tokenlist, TokenList([]))

    def test_append(self):
        tokenlist = TokenList()
        tokenlist.append(Token({"id": 1}))
        self.assertEqual(tokenlist, TokenList([{"id": 1}]))

        tokenlist.append({"id": 2})
        self.assertEqual(tokenlist, TokenList([{"id": 1}, {"id": 2}]))
        self.assertEqual(type(tokenlist[1]), Token)

    def test_insert(self):
        tokenlist = TokenList()
        tokenlist.insert(0, Token({"id": 1}))
        self.assertEqual(tokenlist, TokenList([{"id": 1}]))

        tokenlist.insert(1, {"id": 2})
        self.assertEqual(tokenlist, TokenList([{"id": 1}, {"id": 2}]))
        self.assertEqual(type(tokenlist[1]), Token)

    def test_remove(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}])
        tokenlist.remove(Token({"id": 1}))
        self.assertEqual(tokenlist, TokenList([{"id": 2}]))

        tokenlist.remove({"id": 2})
        self.assertEqual(tokenlist, TokenList())

    def test_index(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}])
        self.assertEqual(tokenlist.index(Token({"id": 1})), 0)
        self.assertEqual(tokenlist.index({"id": 2}), 1)

    def test_count(self):
        tokenlist = TokenList([{"id": 1}, {"id": 1}, {"id": 2}])
        self.assertEqual(tokenlist.count(Token({"id": 1})), 2)
        self.assertEqual(tokenlist.count({"id": 2}), 1)

    def test_setitem_write_token_to_index(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}])
        tokenlist[0] = Token({"id": 3})
        self.assertEqual(tokenlist, TokenList([{"id": 3}, {"id": 2}]))

    def test_setitem_write_dict_to_index(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}])
        tokenlist[1] = {"id": 3}
        self.assertEqual(tokenlist, TokenList([{"id": 1}, {"id": 3}]))
        self.assertEqual(type(tokenlist[1]), Token)

    def test_setitem_write_list_of_dicts_to_slice(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}])
        tokenlist[:] = [{"id": 1}]
        self.assertEqual(tokenlist, TokenList([{"id": 1}]))
        self.assertEqual(type(tokenlist[0]), Token)

    def test_getitem_read_from_slice(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}, {"id": 3}])
        self.assertEqual(tokenlist[0:2], [{"id": 1}, {"id": 2}])

    def test_setitem_write_to_token(self):
        tokenlist = TokenList([{"id": 1}, {"id": 2}])
        self.assertEqual(tokenlist[1]["id"], 2)
        tokenlist[1]["id"] = 3
        self.assertEqual(tokenlist[1]["id"], 3)

class TestParsingTrickyTrees(unittest.TestCase):
    def assertTreeEqual(self, tree, other):
        self.assertEqual(tree.token, other.token)
        self.assertEqual(len(tree.children), len(other.children))
        for i, child in enumerate(tree.children):
            self.assertTreeEqual(child, other.children[i])

    def test_simple_tree(self):
        tokenlist = TokenList([
            Token([("id", 2), ("form", "dog"), ("head", 0)]),
            Token([("id", 1), ("form", "a"), ("head", 2)]),
        ])
        tree = TokenTree(
            token=Token([("id", 2), ("form", "dog"), ("head", 0)]),
            children=[TokenTree(
                token=Token([("id", 1), ("form", "a"), ("head", 2)]),
                children=[]
            )]
        )
        self.assertTreeEqual(tokenlist.to_tree(), tree)

    def test_removes_negative_nodes(self):
        tokenlist = TokenList([
            Token([("id", 2), ("form", "dog"), ("head", 0)]),
            Token([("id", 1), ("form", "a"), ("head", 2)]),
            Token([("id", 3), ("form", "😍"), ("head", -1)]),
        ])
        tree = TokenTree(
            token=Token([("id", 2), ("form", "dog"), ("head", 0)]),
            children=[TokenTree(
                token=Token([("id", 1), ("form", "a"), ("head", 2)]),
                children=[]
            )]
        )
        self.assertTreeEqual(tokenlist.to_tree(), tree)

    def test_multiple_root_nodes(self):
        tokenlist = TokenList([
            Token([('id', 1), ('form', 'To'), ('head', 0)]),
            Token([('id', 2), ('form', 'appear'), ('head', 1)]),
            Token([('id', 4), ('form', 'EMNLP'), ('head', 0)]),
            Token([('id', 5), ('form', '2014'), ('head', 4)]),
            Token([('id', 6), ('form', 'Yay!'), ('head', 0)]),
        ])
        tree = TokenTree(
            token=Token([("id", 0), ("form", "_"), ("deprel", "root")]),
            children=[
                TokenTree(
                    token=Token([("id", 1), ("form", "To"), ("head", 0)]),
                    children=[TokenTree(
                        token=Token([("id", 2), ("form", "appear"), ("head", 1)]),
                        children=[]
                    )]
                ),
                TokenTree(
                    token=Token([("id", 4), ("form", "EMNLP"), ("head", 0)]),
                    children=[TokenTree(
                        token=Token([("id", 5), ("form", "2014"), ("head", 4)]),
                        children=[]
                    )]
                ),
                TokenTree(
                    token=Token([("id", 6), ("form", "Yay!"), ("head", 0)]),
                    children=[]
                ),
            ]
        )
        self.assertTreeEqual(tokenlist.to_tree(), tree)

    def test_no_root_nodes(self):
        tokenlist = TokenList([
            Token([('id', 1), ('form', 'To'), ('head', 1)]),
            Token([('id', 2), ('form', 'appear'), ('head', 2)]),
        ])
        with self.assertRaises(ParseException):
            tokenlist.to_tree()


class TestSerialize(unittest.TestCase):
    def test_serialize_on_tokenlist(self):
        tokenlist = TokenList([{"id": 1}])
        self.assertEqual(tokenlist.serialize(), serialize(tokenlist))


class TestFilter(unittest.TestCase):
    def test_basic_filtering(self):
        tokenlist = TokenList([
            {"id": 1, "form": "a", "field": "x"},
            {"id": 2, "form": "dog", "field": "x"},
        ])
        self.assertEqual(
            tokenlist.filter(id=0),
            TokenList([])
        )
        self.assertEqual(
            tokenlist.filter(id=1),
            TokenList([{"id": 1, "form": "a", "field": "x"}])
        )
        self.assertEqual(
            tokenlist.filter(),
            tokenlist
        )
        self.assertEqual(
            tokenlist.filter(field="x"),
            tokenlist
        )

    def test_and_filtering(self):
        tokenlist = TokenList([
            {"id": 1, "form": "a", "field": "x"},
            {"id": 2, "form": "dog", "field": "x"},
            {"id": 3, "form": "dog", "field": "y"},
        ])
        self.assertEqual(
            tokenlist.filter(field="x", id=2),
            TokenList([
                {"id": 2, "form": "dog", "field": "x"},
            ])
        )
        self.assertEqual(
            tokenlist.filter(field="x", id=3),
            TokenList([])
        )

    def test_deep_filtering(self):
        tokenlist = TokenList([
            {"form": "The", "feats": Token([('Definite', 'Def'), ('PronType', 'Art')])},
            {"form": "quick", "feats": Token([('Degree', 'Pos')])},
            {"form": "brown", "feats": Token([('Degree', 'Pos')])},
            {"form": "fox", "feats": Token([('Number', 'Sing')])},
        ])
        self.assertEqual(
            tokenlist.filter(feats__Degree="Pos"),
            TokenList([
                {"form": "quick", "feats": Token([('Degree', 'Pos')])},
                {"form": "brown", "feats": Token([('Degree', 'Pos')])},
            ])
        )
        self.assertEqual(
            tokenlist.filter(form="brown", feats__Degree="Pos"),
            TokenList([
                {"form": "brown", "feats": Token([('Degree', 'Pos')])},
            ])
        )
        self.assertEqual(
            tokenlist.filter(form="brown", feats__Degree="Pos", id=1),
            TokenList([])
        )
        self.assertEqual(
            tokenlist.filter(unknown__property__value="undefined"),
            TokenList([])
        )
        self.assertEqual(
            tokenlist.filter(unknown___property____value="undefined"),
            TokenList([])
        )

    def test_nested_filtering(self):
        tokenlist = TokenList([
            {"form": "The", "feats": Token([('Definite', 'Def'), ('PronType', 'Art')])},
            {"form": "quick", "feats": Token([('Degree', 'Pos')])},
            {"form": "brown", "feats": Token([('Degree', 'Pos')])},
            {"form": "fox", "feats": Token([('Number', 'Sing')])},
        ])
        self.assertEqual(
            tokenlist.filter(feats__Degree="Pos").filter(form="brown"),
            TokenList([
                {"form": "brown", "feats": Token([('Degree', 'Pos')])},
            ])
        )
        self.assertEqual(
            tokenlist.filter(form="brown").filter(feats__Degree="Pos"),
            TokenList([
                {"form": "brown", "feats": Token([('Degree', 'Pos')])},
            ])
        )
        self.assertEqual(
            tokenlist.filter(form="brown").filter(feats__Degree="Pos").filter(),
            TokenList([
                {"form": "brown", "feats": Token([('Degree', 'Pos')])},
            ])
        )
        self.assertEqual(
            tokenlist.filter(form="brown").filter(feats__Degree="Pos").filter(id=0),
            TokenList([])
        )

    def test_lambda_basic_filtering(self):
        tokenlist = TokenList([
            Token({'id': (1, '-', 2), 'form': "It's", 'lemma': '_', 'feats': None}),
            Token({'id': 1, 'form': 'It', 'lemma': 'it'}),
            Token({'id': 2, 'form': "'s", 'lemma': 'be'})
        ])

        self.assertEqual(
            tokenlist.filter(id=lambda x: type(x) is int),
            TokenList([
                Token({'id': 1, 'form': 'It', 'lemma': 'it'}),
                Token({'id': 2, 'form': "'s", 'lemma': 'be'})
            ])
        )
        self.assertEqual(
            tokenlist.filter(lemma=lambda x: x.startswith('b')),
            TokenList([
                Token({'id': 2, 'form': "'s", 'lemma': 'be'})
            ])
        )

    def test_lambda_deep_filtering(self):
        tokenlist = TokenList([
            Token({'id': (1, '-', 2), 'feats': None}),
            Token({'id': 1, 'feats': {'Case': 'Nom', 'Number': 'Sing'}}),
            Token({'id': 2, 'feats': {'Mood': 'Ind', 'Number': 'Sing'}})
        ])

        self.assertEqual(
            tokenlist.filter(feats__Mood=lambda x: x == 'Ind'),
            TokenList([
                Token({'id': 2, 'feats': {'Mood': 'Ind', 'Number': 'Sing'}})
            ])
        )

        self.assertEqual(
            tokenlist.filter(feats__Number=lambda x: x == 'Sing'),
            TokenList([
                Token({'id': 1, 'feats': {'Case': 'Nom', 'Number': 'Sing'}}),
                Token({'id': 2, 'feats': {'Mood': 'Ind', 'Number': 'Sing'}})
            ])
        )


class TestTokenTree(unittest.TestCase):
    def test_eq(self):
        metadata = {"meta": "data"}

        tokentree1 = TokenTree(token={"id": 1}, children=[TokenTree(token={"id": 2}, children=[])])
        tokentree1.metadata = metadata

        tokentree2 = TokenTree(token={"id": 1}, children=[])
        self.assertNotEqual(tokentree1, tokentree2)

        # Not equal with a plain dict
        self.assertNotEqual(tokentree2, {"id": 1})

        tokentree2.metadata = metadata
        self.assertNotEqual(tokentree1, tokentree2)

        tokentree2.children = [TokenTree(token={"id": 2}, children=[])]
        self.assertEqual(tokentree1, tokentree2)

    def test_metadata(self):
        tree = TokenTree(token={"id": 1, "form": "hej"}, children=[])
        metadata = {"meta": "data"}
        tree.set_metadata(metadata)
        self.assertEqual(tree.metadata, metadata)

        tree = TokenTree(token={"id": 1, "form": "hej"}, children=[], metadata={"meta": "data"})
        self.assertEqual(tree.metadata, metadata)


class TestSerializeTree(unittest.TestCase):
    def test_missing_id(self):
        tree = TokenTree(token={"form": "hej"}, children=[])
        with self.assertRaises(ParseException):
            tree.serialize()

    def test_flatten(self):
        tree = TokenTree(
            token=Token([("id", 2), ("form", "dog")]),
            children=[TokenTree(
                token=Token([("id", 1), ("form", "a")]),
                children=[]
            )]
        )
        self.assertEqual(
            tree.serialize(),
            dedent("""\
                1\ta
                2\tdog

            """)
        )
        tree = TokenTree(
            token=Token([("id", 1), ("form", "dog")]),
            children=[TokenTree(
                token=Token([("id", 2), ("form", "a")]),
                children=[]
            )]
        )
        self.assertEqual(
            tree.serialize(),
            dedent("""\
                1\tdog
                2\ta

            """)
        )


class TestPrintTree(unittest.TestCase):
    def test_print_empty_list(self):
        tree = TokenTree(None, [])
        with self.assertRaises(ParseException):
            capture_print(tree.print_tree)

    def test_tree_without_deprel(self):
        tree = TokenTree(token={"id": 1, "form": "hej"}, children=[])
        with self.assertRaises(ParseException):
            capture_print(tree.print_tree)

    def test_tree_without_id(self):
        tree = TokenTree(token={"form": "hej", "deprel": "nmod"}, children=[])
        with self.assertRaises(ParseException):
            capture_print(tree.print_tree)

    def test_print_simple(self):
        tree = TokenTree(token={"id": "X", "deprel": "Y", "test": "data"}, children=[])
        result = capture_print(tree.print_tree)
        self.assertEqual(result, "(deprel:Y) test:data [X]\n")

    def test_print_with_children(self):
        tree = TokenTree(token={"id": "X", "deprel": "Y", "test": "data"}, children=[
            TokenTree(token={"id": "X", "deprel": "Y", "test": "data"}, children=[]),
            TokenTree(token={"id": "X", "deprel": "Y", "test": "data"}, children=[]),
        ])
        result = capture_print(tree.print_tree)

        self.assertEqual(result, dedent("""\
            (deprel:Y) test:data [X]
                (deprel:Y) test:data [X]
                (deprel:Y) test:data [X]
        """))


class TestSentenceList(unittest.TestCase):
    def test_init_empty_sentences_should_not_raise(self):
        SentenceList()

    def test_init_generator_sentences_should_not_raise(self):
        def test():
            yield

        SentenceList(test())

    def test_init_nonlist_raises(self):
        with self.assertRaises(ParseException):
            SentenceList((1, 2, 4))

    def test_equals(self):
        tokenlists = [TokenList([{"id": 1}])]
        self.assertEqual(SentenceList(tokenlists), SentenceList(tokenlists))
        self.assertEqual(SentenceList(tokenlists), tokenlists)
        self.assertEqual(tokenlists, SentenceList(tokenlists))

        tokenlists2 = [TokenList([{"id": 2}])]
        self.assertNotEqual(SentenceList(tokenlists), SentenceList(tokenlists2))

    def test_clear(self):
        sentences = SentenceList([TokenList([{"id": 1}])], metadata={"a": 1})
        sentences.clear()
        self.assertEqual(sentences, SentenceList())

    def test_copy(self):
        sentences = SentenceList([TokenList([{"id": 1}])], metadata={"a": 1})
        sentences2 = sentences.copy()
        self.assertNotEqual(id(sentences), id(sentences2))
        self.assertEqual(sentences, sentences2)

    def test_extend(self):
        sentences = SentenceList([TokenList([{"id": 1}])], metadata={"a": 1})
        sentences2 = SentenceList([TokenList([{"id": 2}])], metadata={"b": 2})
        sentences3 = [TokenList([{"id": 3}])]

        sentences.extend(sentences2)
        self.assertEqual(
            sentences,
            SentenceList(
                [TokenList([{"id": 1}]), TokenList([{"id": 2}])], metadata={"a": 1, "b": 2}
            )
        )

        sentences.extend(sentences3)
        self.assertEqual(
            sentences,
            SentenceList(
                [
                    TokenList([{"id": 1}]), TokenList([{"id": 2}]), TokenList([{"id": 3}])
                ], metadata={"a": 1, "b": 2}
            )
        )

    def test_slicing_sentencelist_works_and_retains_metadata(self):
        sentences = SentenceList(
            [
                TokenList([{"id": 1}]),
                TokenList([{"id": 2}]),
                TokenList([{"id": 3}]),
            ],
            metadata={"a": 1}
        )
        self.assertEqual(sentences[1:2], SentenceList(
            [
                TokenList([{"id": 2}]),
            ],
            metadata={"a": 1}
        ))

    def test_slicing_with_single_int(self):
        sentences = SentenceList(
            [
                TokenList([{"id": 1}]),
                TokenList([{"id": 2}]),
                TokenList([{"id": 3}]),
            ],
            metadata={"a": 1}
        )
        self.assertEqual(sentences[1], TokenList([{"id": 2}]))

class TestSentenceGenerator(unittest.TestCase):
    def test_init_empty_sentences_should_not_raise(self):
        with self.assertRaises(TypeError):
            SentenceGenerator()

    def test_init_with_list_casts_to_iterator(self):
        sentences = SentenceGenerator([TokenList([{"id": 1}])])
        self.assertEqual(next(sentences), TokenList([{"id": 1}]))

    def test_init_raises_with_non_iterator(self):
        with self.assertRaises(ParseException):
            SentenceGenerator((1, 2, 4))

    def test_generator_generates(self):
        tokenlist1 = TokenList([{"id": 1}])
        tokenlist2 = TokenList([{"id": 2}])

        def two_sentences():
            yield tokenlist1
            yield tokenlist2

        self.assertEqual(list(SentenceGenerator(two_sentences())), [tokenlist1, tokenlist2])
        self.assertEqual(next(SentenceGenerator(two_sentences())), tokenlist1)
